"""Entropies of some useful models. """

from dataclasses import dataclass
from functools import partial

import numpy as np
from typing import Tuple, Optional, Callable

from cupid_utils import bs_error_abort, _EPS, _TWO_EPS

EntropyGradient = Callable[[np.ndarray, np.ndarray, np.ndarray], np.ndarray]
"""The type of a function that takes :math:`(\\mu,n,m)` and returns an array with the gradient of entropy wrt :math:`\\mu`.
"""

EntropyHessianComponents = Tuple[np.ndarray, np.ndarray, np.ndarray]
"""The type of a tuple of the three components of the hessian of the entropy wrt :math:`\\mu`.
"""

EntropyHessian = Callable[[np.ndarray, np.ndarray, np.ndarray], EntropyHessianComponents]
"""The type of a function that takes :math:`(\\mu,n,m)` and returns the three components of the hessian of the entropy wrt :math:`\\mu`.
"""


@dataclass
class EntropyFunctions:
    """Defines the entropy used.

    e0_fun: required

    parameter_dependent:  if ``True``, the entropy depends on parameters. Defaults to ``False``

    e_fun: only in entropies that depend on parameters. Defaults to ``None``

    hessian: defaults to "numeric"
     * if "provide", we provide the hessian of the entropy.
     * if "numerical", it is computed by central differences.

    e0_derivative: the derivative of ``e0_fun``, if available. Defaults to ``None``

    e_derivative: the derivative of ``e_fun``, if available.  Defaults to ``None``

    description: some text describing the model. Defaults to ``None``

    Examples:
        See `EntropyChooSiow` and `EntropyChooSiowHeteroskedastic`
    """
    e0_fun: EntropyGradient
    parameter_dependent: Optional[bool] = False
    e_fun: Optional[EntropyGradient] = None
    hessian: Optional[str] = "numerical"
    e0_derivative: Optional[EntropyHessian] = None
    e_derivative: Optional[EntropyHessian] = None
    description: Optional[str] = None

    def __post_init__(self):
        if not self.parameter_dependent:
            if self.hessian == "provided" and self.e0_derivative is None:
                bs_error_abort("You claim to provide the hessian but you did not provide the e0_derivative.")
        else:
            if self.e_fun is None:
                bs_error_abort("Your entropy is parameter dependent but you did not provide the e_fun.")
            if self.hessian == "provided" and self.e_derivative is None:
                bs_error_abort(
                    "Your entropy is parameter dependent, you claim to provide the hessian,\n but I do not see the e_derivative.")


def entropy_gradient(entropy: EntropyFunctions, muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray,
                     alpha: Optional[np.ndarray] = None) -> np.ndarray:
    """Computes the derivative of the entropy wrt :math:`\\mu` at :math:`\\mu, n, m, \\alpha)`

    Args:
        entropy: the ``EntropyFunctions`` object
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins
        alpha: a vector of parameters of the derivative of the entropy, if any

    Returns:
        the derivative of the entropy wrt :math:`\\mu` at :math:`(\\mu, n, m, \\alpha)`.
    """
    e0_fun = entropy.e0_fun
    e0_vals = e0_fun(muxyhat, n, m)
    parameter_dependent = entropy.parameter_dependent
    if parameter_dependent:
        if alpha is None:
            bs_error_abort("alpha should be specified for this model")
        e_fun = entropy.e_fun
        return (e0_vals + e_fun(muxyhat, n, m) @ alpha)
    else:
        return e0_vals


def _numeric_hessian(entropy: EntropyFunctions, muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray,
                     alpha: Optional[np.ndarray] = None) -> EntropyHessianComponents:
    """Evaluates numerically the three components of the hessian of the entropy wrt :math:`\\mu` at :math:`\\mu, n, m, \\alpha)`

    Args:
        entropy: the ``EntropyFunctions`` object
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins
        alpha: a vector of parameters of the deruvate of the entropy, if any

    Returns:
        the hessian of the entropy wrt :math:`\\mu` at :math:`\\mu, n, m, \\alpha)`.
    """
    parameter_dependent = entropy.parameter_dependent
    if not parameter_dependent:
        entropy_deriv = entropy.e0_fun
    else:
        entropy_deriv = partial(entropy_gradient, entropy, alpha=alpha)

    X, Y = muxyhat.shape
    hessian_x = np.zeros((X, Y, Y))
    hessian_y = np.zeros((X, Y, X))
    hessian_xy = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            for t in range(Y):
                muxy = muxyhat.copy().astype(float)
                muxy[x, t] += _EPS
                der_entropy_plus = entropy_deriv(muxy, n, m)
                muxy[x, t] -= _TWO_EPS
                der_entropy_minus = entropy_deriv(muxy, n, m)
                hessian_x[x, y, t] = (der_entropy_plus[x, y] - der_entropy_minus[x, y]) / _TWO_EPS
            for z in range(X):
                muxy = muxyhat.copy().astype(float)
                muxy[z, y] += _EPS
                der_entropy_plus = entropy_deriv(muxy, n, m)
                muxy[z, y] -= _TWO_EPS
                der_entropy_minus = entropy_deriv(muxy, n, m)
                hessian_y[x, y, z] = (der_entropy_plus[x, y] - der_entropy_minus[x, y]) / _TWO_EPS
            muxy = muxyhat.copy().astype(float)
            muxy[x, y] += _EPS
            der_entropy_plus = entropy_deriv(muxy, n, m)
            muxy[x, y] -= _TWO_EPS
            der_entropy_minus = entropy_deriv(muxy, n, m)
            hessian_xy[x, y] = (der_entropy_plus[x, y] - der_entropy_minus[x, y]) / _TWO_EPS

    return hessian_x, hessian_y, hessian_xy


def _fill_hessian_from_components(hessian_components: EntropyHessianComponents) -> np.ndarray:
    """Fills the hessian of the entropy wrt :math:`\\mu`

    Args:
        hessian_components: the three components of the hessian

    Returns:
        the (XY,XY) matrix of the hessian
    """
    hess_x, hess_y, hess_xy = hessian_components
    X, Y = hess_xy.shape
    XY = X * Y
    hessian = np.zeros((XY, XY))

    i = 0
    ix = 0
    for x in range(X):
        for y in range(Y):
            hessian[i, ix:(ix + Y)] = hess_x[x, y, :]
            slice_y = slice(y, XY, Y)
            hessian[i, slice_y] = hess_y[x, y, :]
            hessian[i, i] = hess_xy[x, y]
            i += 1
        ix += Y

    return hessian

