""" Estimates semilinear separable models with a given entropy function.
The entropy function and the surplus matrix must both be linear in the parameters.
"""
import scipy.linalg as spla
import scipy.stats as sts
from typing import Optional

from cupid_utils import test_vector, test_matrix, bs_error_abort, print_stars, _make_phi_mat
from entropy import EntropyFunctions, _numeric_hessian, _fill_hessian_from_components
from choo_siow import entropy_choo_siow, entropy_choo_siow_numeric, simulate_choo_siow_market
from choo_siow_heteroskedastic import entropy_choo_siow_heteroskedastic, \
    entropy_choo_siow_heteroskedastic_numeric
from choo_siow_gender_heteroskedastic import entropy_choo_siow_gender_heteroskedastic, \
    entropy_choo_siow_gender_heteroskedastic_numeric
from min_distance_utils import MDEResults, _variance_muxyhat, _compute_estimates
from nests_parameters import *
from two_layer_nested_logit import entropy_nested_logit, entropy_nested_logit_numeric


def estimate_semilinear_mde(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray,
                            phi_bases: np.ndarray,
                            entropy: EntropyFunctions,
                            initial_omega: Optional[np.ndarray] = None) -> MDEResults:
    """
    Estimates the parameters of the distributions and of the base functions.

    Args:
        muxyhat: an (X, Y) matrix of observed matching patterns
        n: an X-vector of observed margins
        m: an Y-vector of observed margins
        phi_bases: an (X, Y, K) array of bases
        entropy: an ``EntropyFunctions`` object
        initial_omega: if specified, used as the weighting matrix for the first step when `param_entropy` is `True`

    Returns:
        an `MDEResults` instance

    Example:
        See the `main` function in the source of the module.
    """
    X, Y = test_matrix(muxyhat)
    Xn = test_vector(n)
    if Xn != X:
        bs_error_abort(f"n has {Xn} elements while muxyhat is ({X}, {Y})")
    Ym = test_vector(m)
    if Ym != Y:
        bs_error_abort(f"m has {Ym} elements while muxyhat is ({X}, {Y})")
    ndims_phi = phi_bases.ndim
    if ndims_phi != 3:
        bs_error_abort(f"phi_bases should have 3 dimensions, not {ndims_phi}")
    Xp, Yp, K = phi_bases.shape
    if Xp != X or Yp != Y:
        bs_error_abort(
            f"phi_bases should have shape ({X}, {Y}, {K}) not ({Xp}, {Yp}, {K})")
    parameterized_entropy = entropy.parameter_dependent
    if parameterized_entropy and initial_omega is None:
        print_stars("Using the identity matrix as weighting matrix in the first step.")
        Omega = np.eye(X * Y)

    phi_mat = _make_phi_mat(phi_bases)
    e0_fun = entropy.e0_fun
    e0_hat = e0_fun(muxyhat, n, m).ravel()

    if not parameterized_entropy:
        n_pars = K
        hessian = entropy.hessian
        if hessian == "provided":
            e0_derivative = entropy.e0_derivative
            der2_entropy_hat = e0_derivative(muxyhat, n, m)
        else:
            der2_entropy_hat = _numeric_hessian(entropy, muxyhat, n, m)
        der2_entropy_hat = _fill_hessian_from_components(der2_entropy_hat)

        var_muxy = _variance_muxyhat(muxyhat, n, m)
        var_entropy_gradient = der2_entropy_hat.T @ var_muxy @ der2_entropy_hat
        # print(spla.eigvals(var_entropy_gradient))
        Omega = spla.inv(var_entropy_gradient)
        estimated_coefficients, varcov_coefficients = _compute_estimates(phi_mat, Omega, e0_hat)
        stderrs_coefficients = np.sqrt(np.diag(varcov_coefficients))
        est_Phi = phi_mat @ estimated_coefficients
        residuals = est_Phi + e0_hat
    else:  # parameterized entropy
        # create the F matrix
        e_fun = entropy.e_fun
        e_hat = _make_phi_mat(e_fun(muxyhat, n, m))

        F_hat = np.column_stack((e_hat, phi_mat))
        F_T = F_hat.T
        n_pars = e_hat.shape[1] + K
        # first pass with an initial Omega
        first_coeffs, _ = _compute_estimates(F_hat, Omega, e0_hat)
        first_alpha = first_coeffs[:-K]

        # compute the efficient Omega
        hessian = entropy.hessian
        if hessian == "provided":
            e0_derivative = entropy.e0_derivative
            e_derivative = entropy.e_derivative
            der2_e0_hat = e0_derivative(muxyhat, n, m)
            der2_e_hat = e_derivative(muxyhat, n, m)

            print_stars("First-stage estimates:")
            print(first_coeffs)

            first_alpha = first_coeffs[:-K]

            der2_entropy_hat = []
            for c in [0, 1, 2]:
                der2_entropy_hat.append(der2_e0_hat[c] + der2_e_hat[c] @ first_alpha)
            der2_entropy_hat = tuple(der2_entropy_hat)
        else:
            der2_entropy_hat = _numeric_hessian(entropy, muxyhat, n, m, alpha=first_alpha)
        der2_entropy_hat = _fill_hessian_from_components(der2_entropy_hat)

        var_muxy = _variance_muxyhat(muxyhat, n, m)
        var_entropy_gradient = der2_entropy_hat.T @ var_muxy @ der2_entropy_hat
        Omega = spla.inv(var_entropy_gradient)

        # second pass
        estimated_coefficients, varcov_coefficients = _compute_estimates(F_hat, Omega, e0_hat)
        est_alpha, est_beta = estimated_coefficients[:-K], estimated_coefficients[-K:]
        stderrs_coefficients = np.sqrt(np.diag(varcov_coefficients))
        est_Phi = phi_mat @ est_beta
        residuals = est_Phi + e0_hat + e_hat @ est_alpha

    value_obj = residuals.T @ Omega @ residuals
    ndf = X * Y - n_pars
    test_stat = value_obj
    n_households = np.sum(n) + np.sum(m) - np.sum(muxyhat)

    results = MDEResults(X=X, Y=Y, K=K, number_households=n_households,
                         estimated_coefficients=estimated_coefficients,
                         varcov_coefficients=varcov_coefficients,
                         stderrs_coefficients=stderrs_coefficients,
                         estimated_Phi=est_Phi.reshape((X, Y)),
                         test_statistic=test_stat,
                         ndf=ndf,
                         test_pvalue=sts.chi2.sf(test_stat, ndf),
                         parameterized_entropy=parameterized_entropy)
    return results


if __name__ == '__main__':

    X, Y, K = 5, 6, 2
    n_indivs = 1e5
    muxy_sim, n_sim, m_sim, phi_bases, lambda_true = simulate_choo_siow_market(X, Y, K, n_indivs)

    entropy_models = [entropy_choo_siow, entropy_choo_siow_numeric,
                      entropy_choo_siow_gender_heteroskedastic,
                      entropy_choo_siow_gender_heteroskedastic_numeric,
                      entropy_choo_siow_heteroskedastic,
                      entropy_choo_siow_heteroskedastic_numeric,
                      entropy_nested_logit,
                      entropy_nested_logit_numeric
                      ]

    for entropy_model in entropy_models:
        if entropy_model.description is not None:
            print_stars(entropy_model.description)

        mde_results = estimate_semilinear_mde(muxy_sim, n_sim, m_sim, phi_bases,
                                              entropy_model)
        estimates = mde_results.estimated_coefficients

        if entropy_model == entropy_choo_siow_gender_heteroskedastic:
            true_alpha = np.ones(1)

        if entropy_model == entropy_choo_siow_heteroskedastic:
            true_alpha = np.ones(estimates.size - lambda_true.size)

        if entropy_model in [entropy_nested_logit, entropy_nested_logit_numeric]:
            true_alpha = np.zeros(4)
            if n_X != X or n_Y != Y:
                bs_error_abort(f"muxy has shape ({X}, {Y}) but nests_parameters has ({n_X}, {n_Y})")


        if entropy_model.parameter_dependent:
            true_coeffs = np.concatenate((true_alpha, lambda_true))
        else:
            true_coeffs = lambda_true

        print_stars("True and estimated coefficients")
        print(np.column_stack((true_coeffs, estimates)))

        print_stars(
            f"Test statistic: {mde_results.test_statistic} has p-value {mde_results.test_pvalue}")

        print(mde_results)
