"""Utility programs used in `min_distance.py`.
"""
from dataclasses import dataclass
import numpy as np
import scipy.linalg as spla
from typing import Optional, Union


@dataclass
class MDEResults:
    """The results from estimation and testing.    """
    X: int
    Y: int
    K: int
    number_households: int
    estimated_coefficients: np.ndarray
    varcov_coefficients: np.ndarray
    stderrs_coefficients: np.ndarray
    estimated_Phi: np.ndarray
    test_statistic: float
    test_pvalue: float
    ndf: int
    parameterized_entropy: Optional[bool] = False

    def __repr__(self):
        line_stars = '*' * 80 + "\n"
        if self.parameterized_entropy:
            n_alpha = self.estimated_coefficients.size - self.K
            entropy_str = f"     The entropy has {n_alpha} parameters."
        else:
            entropy_str= "     The entropy is parameter-free."
            n_alpha = 0
        model_str = f"The data has {self.number_households} households\n\n"
        model_str += f"The model has {self.X}x{self.Y} margins\n {entropy_str} \n"
        model_str += f"We use {self.K} basis functions.\n\n"
        repr_str = line_stars + model_str
        repr_str += "The estimated coefficients (and their standard errors) are\n\n"
        if self.parameterized_entropy:
            for i, coeff in enumerate(self.estimated_coefficients[:n_alpha]):
                repr_str += f"   alpha({i + 1}): {coeff: > 10.3f}  ({self.stderrs_coefficients[i]: .3f})\n"
            repr_str += "\n"
        for i, coeff in enumerate(self.estimated_coefficients[n_alpha:]):
            repr_str += f"   base {i + 1}: {coeff: > 10.3f}  ({self.stderrs_coefficients[n_alpha+i]: .3f})\n"
        repr_str += "\nSpecification test:\n"
        repr_str += f"   the value of the test statistic is {self.test_statistic: > 10.3f}\n"
        repr_str += f"     for a chi2({self.ndf}), the p-value is {self.test_pvalue: > 10.3f}\n"
        return repr_str + line_stars


def _compute_estimates(M: np.ndarray, Omega: np.ndarray, d: np.ndarray):
    """Returns the QGLS estimates and their variance-covariance.

    Args:
        M: an (XY,p) matrix
        Omega: an (XY, XY) weighting matrix
        d: an XY-vector

    Returns:
        the p-vector of estimates and their estimated (p,p) variance
    """
    M_T = M.T
    M_omega_d = M_T @ Omega @ d
    M_omega_M = M_T @ Omega @ M
    est_coeffs = -spla.solve(M_omega_M, M_omega_d)
    varcov_coeffs = spla.inv(M_omega_M)
    return est_coeffs, varcov_coeffs

def _variance_muxyhat(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray) -> np.ndarray:
    """
    Computes the unweighted variance-covariance matrix of the observed matching patterns

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins

    Returns:
        an (XY, XY) matrix of the variance-covariance of the matching patterns
    """
    n_households = np.sum(n) + np.sum(m) - np.sum(muxyhat)
    mu_norm = (muxyhat / n_households).ravel()
    var_norm = np.diag(mu_norm) - np.outer(mu_norm, mu_norm)
    return n_households * var_norm

