"""Estimates the semilinear Choo and Siow homoskedastic (2006) model using Poisson GLM.
"""


import numpy as np
import scipy.sparse as spr
from sklearn import linear_model
from typing import Optional
from dataclasses import dataclass

from choo_siow import simulate_choo_siow_market
from cupid_utils import _get_singles, print_stars, _make_phi_mat, _check_dimensions


@dataclass
class PoissonGLMResults:
    """Stores and formats the estimation results.    """
    X: int
    Y: int
    K: int
    number_households: int
    number_individuals: int
    estimated_coefficients: np.ndarray
    estimated_Phi: np.ndarray
    estimated_u: np.ndarray
    estimated_v: np.ndarray

    # varcov_coefficients: np.ndarray
    # stderrs_coefficients: np.ndarray

    def __repr__(self):
        line_stars = '*' * 80 + "\n"
        print_stars("Estimating a Choo and Siow model by Poisson GLM.")
        model_str = f"The data has {self.number_households} households\n\n"
        model_str += f"We use {self.K} basis functions.\n\n"
        repr_str = line_stars + model_str
        repr_str += "The estimated coefficients (and their standard errors, some day) are\n\n"
        for i, coeff in enumerate(self.estimated_coefficients):
            # repr_str += f"   base {i + 1}: {coeff: > 10.3f}  ({self.stderrs_coefficients[i]: .3f})\n"
            repr_str += f"   base {i + 1}: {coeff: > 10.3f}\n"
        repr_str += "The estimated utilities of men are\n\n"
        for i, u_val in enumerate(self.estimated_u):
            repr_str += f"   base {i + 1}: {u_val: > 10.3f}\n"
        repr_str += "The estimated utilities of men are\n\n"
        for i, v_val in enumerate(self.estimated_v):
            repr_str += f"   base {i + 1}: {v_val: > 10.3f}\n"
        return repr_str + line_stars


def _prepare_data(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray):
    """Normalizes the matching patterns and stacks them.

       Args:
        muxyhat: an (X, Y) matrix of observed matching patterns
        n: an X-vector of observed margins
        m: an Y-vector of observed margins
        phi_bases: an (X, Y, K) array of bases
        entropy: an ``EntropyFunctions`` object
        initial_omega: if specified, used as the weighting matrix for the first step when `param_entropy` is `True`

       Returns:
        an `MDEResults` instance

       Example:
        See the `main` function in the source of the module.
    """
    mux0hat, mu0yhat = _get_singles(muxyhat, n, m)
    n_households = muxyhat.sum() + mux0hat.sum() + mu0yhat.sum()
    n_individuals = n_households + muxyhat.sum()
    # rescale the data so that the total number of individuals is one
    muhat_norm = np.concatenate([muxyhat.flatten(), mux0hat, mu0yhat]) / n_individuals
    return muhat_norm, n_households, n_individuals


def choo_siow_poisson_glm(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray, phi_bases: np.ndarray,
                          tol: Optional[float] = 1e-12):
    """
    Estimates the semilinear Choo and Siow homoskedastic (2006) model using Poisson GLM.

       Args:
        muxyhat: an (X, Y) matrix of observed matching patterns
        n: an X-vector of observed margins
        m: an Y-vector of observed margins
        phi_bases: an (X, Y, K) array of bases
        tol: tolerance level for ``linear_model.PoissonRegressor.fit``

       Returns:
        a `PoissonGLMResults` instance

       Example:
        See the `main` function in the source of the module.
    """
    X, Y, K = _check_dimensions(muxyhat, n, m, phi_bases)

    XY = X * Y
    d_a = np.append(2 * np.ones(XY), np.ones(X + Y))
    B = spr.vstack([
        spr.hstack([
            spr.kron(spr.identity(X), np.ones((Y, 1))), spr.kron(np.ones((X, 1)), spr.identity(Y))
        ]),
        spr.hstack([
            spr.identity(X), spr.csr_matrix((X, Y))
        ]),
        spr.hstack([
            spr.csr_matrix((Y, X)), spr.identity(Y)
        ])
    ])

    muhat_norm, n_households, n_individuals = _prepare_data(muxyhat, n, m)
    phi_mat = np.vstack((_make_phi_mat(phi_bases), np.zeros((X, K)), np.zeros((Y, K))))

    C = spr.hstack([-B, phi_mat])
    clf = linear_model.PoissonRegressor(fit_intercept=False, tol=tol, verbose=3, alpha=0)
    clf.fit(spr.diags(1 / d_a) @ C, muhat_norm, sample_weight=d_a)

    u_est = clf.coef_[:X]
    v_est = clf.coef_[X: -K]
    lambda_est = clf.coef_[-K:]
    Phi_est = phi_bases @ lambda_est

    results = PoissonGLMResults(X=X, Y=Y, K=K, number_households=n_households, number_individuals=n_individuals,
                                estimated_coefficients=lambda_est, estimated_Phi=Phi_est,
                                estimated_u=u_est, estimated_v=v_est)

    return results


if __name__ == '__main__':
    X, Y, K = 5, 6, 2
    n_indivs = 1e5
    muxy_sim, n_sim, m_sim, phi_bases, lambda_true = simulate_choo_siow_market(X, Y, K, n_indivs)

    results = choo_siow_poisson_glm(muxy_sim, n_sim, m_sim, phi_bases)
    print(results)

    print_stars("The true coefficients are:")
    print(lambda_true)
