"""The components of the derivative of the entropy for a two-layer nested logit model.
One nest on each side must consist of the 0 option. The other nests are specified in nested lists in ``nests_parameters.py``.
On each side, the nests are the same for each type, with the same parameters.
"""
from math import log

from entropy import EntropyHessianComponents, EntropyFunctions
from cupid_utils import _get_singles
from nests_parameters import *

def e0_nested_logit(muxyhat: np.ndarray,
                    n: np.ndarray, m: np.ndarray) -> np.ndarray:
    """Returns the values of the parameter-independent part :math:`e_0` at :math:`\\hat{\\mu}`
    for the nested logit.

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins

    Returns:
        the (X,Y) matrix of the parameter-independent part of the first derivative of the entropy.
    """
    muxy = muxyhat.astype(float)
    X, Y = muxy.shape
    mux0, mu0y = _get_singles(muxy, n, m)
    e0_vals = np.zeros((X, Y))

    for x in range(X):
        mux0_x = mux0[x]
        for nest in nests_x_over_y:
            mu_xn = np.sum(muxy[x, nest])
            e0_vals[x, nest] = - log(mu_xn / mux0_x)
    for y in range(Y):
        mu0y_y = mu0y[y]
        for nest in nests_y_over_x:
            mu_ny = np.sum(muxy[nest, y])
            e0_vals[nest, y] -= log(mu_ny / mu0y_y)
    return e0_vals


def e0_derivative_nested_logit(muxyhat: np.ndarray,
                               n: np.ndarray, m: np.ndarray) -> EntropyHessianComponents:
    """Returns the derivatives of the parameter-independent part :math:`e_0` at :math:`\\hat{\\mu}`
    for the nested logit.

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins

    Returns:
        the three components of the parameter-independent part of the hessian of the entropy.
    """
    X, Y = muxyhat.shape
    muxy = muxyhat.astype(float)
    mux0, mu0y = _get_singles(muxy, n, m)

    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    der_logxy = 1.0 / muxy
    der_logx0 = 1.0 / mux0
    der_log0y = 1.0 / mu0y

    for x in range(X):
        dlogx0 = der_logx0[x]
        for nest in nests_x_over_y:
            mu_xn = np.sum(muxy[x, nest])
            der_logxn = 1.0 / mu_xn
            for y in nest:
                hess_x[x, y, :] = - dlogx0
                hess_x[x, y, nest] -= der_logxn
    for y in range(Y):
        dlog0y = der_log0y[y]
        for nest in nests_y_over_x:
            mu_ny = np.sum(muxy[nest, y])
            der_logny = 1.0 / mu_ny
            for x in nest:
                hess_y[x, y, :] = - dlog0y
                hess_y[x, y, nest] -= der_logny
    for x in range(X):
        for y in range(Y):
            hess_xy[x, y] = hess_x[x, y, y] + hess_y[x, y, x]

    return hess_x, hess_y, hess_xy


def e_nested_logit(muxyhat: np.ndarray,
                   n: np.ndarray, m: np.ndarray) -> np.ndarray:
    """Returns the values of the parameter-dependent part  :math:`e` at :math:`\\hat{\\mu}`
    for the nested logit.

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins

    Returns:
        the (X,Y,n_alpha) array of the parameter-dependent part of the first derivative of the entropy.
    """
    X, Y = muxyhat.shape
    muxy = muxyhat.astype(float)

    e_vals = np.zeros((X, Y, n_alpha))

    for x in range(X):
        for i_n, nest in enumerate(nests_x_over_y):
            mux_nest_n = muxy[x, nest]
            mu_xn = np.sum(mux_nest_n)
            e_vals[x, nest, i_n] = -np.log(mux_nest_n / mu_xn)

    for y in range(Y):
        for i_n, nest in enumerate(nests_y_over_x):
            muy_nest_n = muxy[nest, y]
            mu_ny = np.sum(muy_nest_n)
            e_vals[nest, y, (i_n + n_X_nests - 1)] -= np.log(muy_nest_n / mu_ny)

    return e_vals


def e_derivative_nested_logit(muxyhat: np.ndarray,
                              n: np.ndarray, m: np.ndarray) -> EntropyHessianComponents:
    """Returns the derivatives of the parameter-dependent part :math:`e` at :math:`\\hat{\\mu}`
     for the nested logit.

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins

    Returns:
        the three components of the parameter-independent part of the hessian of the entropy.
    """
    X, Y = muxyhat.shape
    muxy = muxyhat.astype(float)

    hess_x = np.zeros((X, Y, Y, n_alpha))
    hess_y = np.zeros((X, Y, X, n_alpha))
    hess_xy = np.zeros((X, Y, n_alpha))
    der_logxy = 1.0 / muxy

    for x in range(X):
        for i_n, nest in enumerate(nests_x_over_y):
            mux_nest_n = muxy[x, nest]
            mu_xn = np.sum(mux_nest_n)
            der_logxn = 1.0 / mu_xn
            for t in nest:
                hess_x[x, nest, t, i_n] = der_logxn
            hess_xy[x, nest, i_n] = der_logxn - der_logxy[x, nest]

    for y in range(Y):
        for i_n, nest in enumerate(nests_y_over_x):
            muy_nest_n = muxy[nest, y]
            mu_ny = np.sum(muy_nest_n)
            der_logny = 1.0 / mu_ny
            i_n2 = i_n + n_X_nests - 1
            for z in nest:
                hess_y[nest, y, z, i_n2] = der_logny
            hess_xy[nest, y, i_n2] = der_logny - der_logxy[nest, y]

    return hess_x, hess_y, hess_xy


nest_description = "      each x has the same nests over 0, 1, ..., Y:\n"
for n in nests_for_each_x:
    nest_description += f"      {n}\n"
nest_description += "      each y has the same nests over 0, 1, ..., X:\n"
for n in nests_for_each_y:
    nest_description += f"      {n}\n"
nest_description += "       the parameters rho do not depend on the type."

entropy_nested_logit = EntropyFunctions(
    e0_fun=e0_nested_logit,
    parameter_dependent=True, e_fun=e_nested_logit,
    hessian="provided",
    e0_derivative=e0_derivative_nested_logit,
    e_derivative=e_derivative_nested_logit,
    description="Two-layer nested logit with analytic Hessian\n" + nest_description)

entropy_nested_logit_numeric = EntropyFunctions(
    e0_fun=e0_nested_logit,
    parameter_dependent=True,
    e_fun=e_nested_logit,
    description="Two-layer nested logit with numerical Hessian\n" + nest_description)
